@@ -63,7 +63,7 @@ use common::sense;
 
 use Carp;
 
-our $VERSION = 0.4;
+our $VERSION = 0.5;
 
 use Scalar::Util ();
 
@@ -88,7 +88,7 @@ sub tolc($) {
    lc
 }
 
-=item $fcp = new AnyEvent::FCP [host => $host][, port => $port][, name => $name]
+=item $fcp = new AnyEvent::FCP key => value...;
 
 Create a new FCP connection to the given host and port (default
 127.0.0.1:9481, or the environment variables C<FREDHOST> and C<FREDPORT>).
@@ -96,6 +96,55 @@ Create a new FCP connection to the given host and port (default
 If no C<name> was specified, then AnyEvent::FCP will generate a
 (hopefully) unique client name for you.
 
+The following keys can be specified (they are all optional):
+
+=over 4
+
+=item name => $string
+
+A unique name to identify this client. If none is specified, a randomly
+generated name will be used.
+
+=item host => $hostname
+
+The hostname or IP address of the freenet node. Default is C<$ENV{FREDHOST}>
+or C<127.0.0.1>.
+
+=item port => $portnumber
+
+The port number of the FCP port. Default is C<$ENV{FREDPORT}> or C<9481>.
+
+=item timeout => $seconds
+
+The timeout, in seconds, after which a connection error is assumed when
+there is no activity. Default is C<7200>, i.e. two hours.
+
+=item keepalive => $seconds
+
+The interval, in seconds, at which keepalive messages will be
+sent. Default is C<540>, i.e. nine minutes.
+
+These keepalive messages are useful both to detect that a connection is
+no longer working and to keep any (home) routers from expiring their
+masquerading entry.
+
+=item on_eof => $callback->($fcp)
+
+Invoked when the underlying L<AnyEvent::Handle> signals EOF, currently
+regardless of whether the EOF was expected or not.
+
+=item on_error => $callback->($fcp, $message)
+
+Invoked on any (fatal) errors, such as unexpected connection close. The
+callback receives the FCP object and a textual error message.
+
+=item on_failure => $callback->($fcp, $type, $backtrace, $args, $error)
+
+Invoked when an FCP request fails that didn't have a failure callback. See
+L<FCP REQUESTS> for details.
+
+=back
+
 =cut
 
 sub new {
@@ -137,7 +186,7 @@ sub new {
             # remember end marker
             $rdata = $1 eq "Data"
                or $1 eq "EndMessage"
-               or die "protocol error, expected message end, got $1\n";
+               or return $self->fatal ("protocol error, expected message end, got $1\n");
 
             my @lines = split /\012/, substr $hdl->{rbuf}, 0, $-[0];
 
@@ -191,12 +240,17 @@ sub new {
       $self->{hdl} = new AnyEvent::Handle
          connect  => [$self->{host} => $self->{port}],
          timeout  => $self->{timeout},
-         on_error => sub {
-            warn "$self->{host}: $_[2]\n";#d#
-            exit 1;
-         },
          on_read  => $on_read,
-         on_eof   => $self->{on_eof} || sub { },
+         on_eof   => sub {
+            if ($self->{on_eof}) {
+               $self->{on_eof}($self);
+            } else {
+               $self->fatal ("EOF");
+            }
+         },
+         on_error => sub {
+            $self->fatal ($_[2]);
+         },
       ;
 
       Scalar::Util::weaken ($self->{hdl}{fcp} = $self);
@@ -210,6 +264,19 @@ sub new {
    $self
 }
 
+sub fatal {
+   my ($self, $msg) = @_;
+
+   $self->{hdl}->shutdown;
+   delete $self->{kw};
+   
+   if ($self->{on_error}) {
+      $self->{on_error}->($self, $msg);
+   } else {
+      die $msg;
+   }
+}
+
 sub identifier {
    $_[0]{prefix} . ++$_[0]{idseq}
 }
@@ -372,11 +439,26 @@ This callback/continuation argument (C<$cb>) can come in three forms itself:
 =item A code reference (or rather anything not matching some other alternative)
 
 This code reference will be invoked with the result on success. On an
-error, it will die (in the event loop) with a backtrace of the call site.
+error, it will invoke the C<on_failure> callback of the FCP object, or,
+if none was defined, will die (in the event loop) with a backtrace of the
+call site.
 
 This is a popular choice, but it makes handling errors hard - make sure
 you never generate protocol errors!
 
+If an C<on_failure> hook exists, it will be invoked with the FCP object,
+the request type (the name of the method), a (textual) backtrace as
+generated by C<Carp::longmess>, and arrayref containing the arguments from
+the original request invocation and the error object from the server, in
+this order, e.g.:
+
+   on_failure => sub {
+      my ($fcp, $request_type, $backtrace, $orig_args, $error_object) = @_;
+
+      warn "FCP failure ($type), $error_object->{code_description} ($error_object->{extra_description})$backtrace";
+      exit 1;
+   },
+
 =item A condvar (as returned by e.g. C<< AnyEvent->condvar >>)
 
 When a condvar is passed, it is sent (C<< $cv->send ($results) >>) the
@@ -390,11 +472,14 @@ This is also a popular choice.
 The C<$success> callback will be invoked with the results, while the
 C<$failure> callback will be invoked on any errors.
 
+The C<$failure> callback will be invoked with the error object from the
+server.
+
 =item C<undef>
 
 This is the same thing as specifying C<sub { }> as callback, i.e. on
-success, the results are ignored, while on failure, you the module dies
-with a backtrace.
+success, the results are ignored, while on failure, the C<on_failure> hook
+is invoked or the module dies with a backtrace.
 
 This is good for quick scripts, or when you really aren't interested in
 the results.
@@ -424,9 +509,15 @@ sub _txn {
       } elsif (UNIVERSAL::isa $ok, AnyEvent::CondVar::) {
          $err = sub { $ok->croak ($_[0]{extra_description}) };
       } else {
-         my $bt = Carp::longmess "";
+         my $bt = Carp::longmess "AnyEvent::FCP request $name";
+         Scalar::Util::weaken (my $self = $_[0]);
+         my $args = [@_]; shift @$args;
          $err = sub {
-            die "$_[0]{code_description} ($_[0]{extra_description})$bt";
+            if ($self->{on_failure}) {
+               $self->{on_failure}($self, $name, $args, $bt, $_[0]);
+            } else {
+               die "$_[0]{code_description} ($_[0]{extra_description})$bt";
+            }
          };
       }
 
